from admin_auto_filters.filters import AutocompleteFilter
from django import forms
from django.contrib import admin, messages
from django.contrib.admin.widgets import (
    AutocompleteSelectMultiple,
)
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.db import models
from django.db.models import JSONField, Q
from django.http import HttpResponseRedirect
from django.shortcuts import redirect, render
from django.urls import path, reverse
from django.utils.timezone import now
from django_admin_relation_links import AdminChangeLinksMixin
from import_export import fields
from import_export.admin import ImportExportModelAdmin
from import_export.resources import ModelResource
from jsoneditor.forms import JSONEditor

from .models import (
    Category,
    Question,
    Response,
    Survey,
    UserSurvey,
    UserSurveyNotification,
)
from .resources import HashIdWidget

User = get_user_model()


class ReadOnlyInline(admin.TabularInline):
    def has_change_permission(self, request, obj=None):
        return False

    def has_add_permission(self, request, obj=None):
        return False

    def has_delete_permission(self, request, obj=None):
        return False


@admin.register(Category)
class CategoryAdmin(admin.ModelAdmin):
    list_display = ("id", "slug")


class QuestionInline(admin.TabularInline):
    formfield_overrides = {models.TextField: {"widget": forms.TextInput()}}
    model = Question
    extra = 0


class SurveyUsersForm(forms.Form):
    users = forms.ModelMultipleChoiceField(
        queryset=User.objects.all(),
        widget=AutocompleteSelectMultiple(
            UserSurvey._meta.get_field("user"),
            admin.site,
        ),
        required=False,
    )
    groups = forms.ModelMultipleChoiceField(
        queryset=Group.objects.all(),
        widget=AutocompleteSelectMultiple(
            User._meta.get_field("groups"),
            admin.site,
        ),
        required=False,
    )

    def save(self, survey):
        users = self.cleaned_data["users"]
        groups = self.cleaned_data["groups"]

        created_user_surveys = []
        for user in User.objects.filter(
            Q(groups__in=groups) | Q(id__in=users.values("id"))
        ):
            user_survey, created = UserSurvey.objects.get_or_create(
                user=user, survey=survey
            )
            if created:
                created_user_surveys.append(user_survey)
        return created_user_surveys


class QuestionResource(ModelResource):
    instances_to_keep = set()
    id = fields.Field(column_name="id", attribute="id", widget=HashIdWidget())

    class Meta:
        model = Question
        fields = ["id", "question", "text", "type", "format"]
        export_order = fields

    def before_import_row(self, row, row_number=None, **kwargs):
        # Clean IDs if we are importing questions from another survey
        if row["id"] not in kwargs["question_ids"]:
            row["id"] = None
        if row["question"] is None:
            row["question"] = ""

    def skip_row(self, instance, original, row, import_validation_errors=None):
        return not instance.question

    def after_import_instance(self, instance, new, row_number=None, **kwargs):
        instance.survey_id = kwargs["survey"].id
        instance.sequence = row_number

    def after_save_instance(self, instance, using_transactions, dry_run):
        self.instances_to_keep.add(instance.id)

    def after_import(self, dataset, result, using_transactions, dry_run, **kwargs):
        # TODO: uncomment if we want to remove questions that are not in the file
        # Question.objects.filter(survey_id=kwargs["survey_id"]).exclude(
        #     id__in=self.instances_to_keep
        # ).delete()
        pass


class QuestionResponseResource(ModelResource):
    class Meta:
        model = Question
        fields = ["question"]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.survey = kwargs["survey"]
        self.fields.update(
            {
                user: fields.Field(column_name=user, attribute=user)
                for user in self.survey.get_respondents()
            }
        )


class QuestionResponseExport(ImportExportModelAdmin):
    model = Question
    export_template_name = "admin/df_survey/survey/export_question_responses.html"
    resource_class = QuestionResponseResource

    def get_export_queryset(self, request):
        return request.kwargs["survey"].get_responses()

    def get_export_resource_kwargs(self, request, *args, **kwargs):
        return {
            **super().get_export_resource_kwargs(request, *args, **kwargs),
            "survey": request.kwargs["survey"],
        }

    def get_export_filename(self, request, queryset, file_format):
        return "%s-%s-%s.%s" % (
            request.kwargs["survey"].title,
            "responses",
            now().strftime("%Y-%m-%d"),
            file_format.get_extension(),
        )


class QuestionImportExport(ImportExportModelAdmin):
    model = Question
    import_template_name = "admin/df_survey/survey/import_questions.html"
    export_template_name = "admin/df_survey/survey/export_questions.html"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.admin_site = admin.site

    resource_class = QuestionResource

    def get_import_data_kwargs(self, request, *args, **kwargs):
        survey = request.kwargs["survey"]
        return {
            **super().get_import_data_kwargs(request, *args, **kwargs),
            "survey": survey,
            "question_ids": set(survey.question_set.all().values_list("id", flat=True)),
        }

    def get_export_queryset(self, request):
        return Question.objects.filter(survey__id=request.kwargs["survey"].id)

    def process_result(self, result, request):
        super().process_result(result, request)
        return HttpResponseRedirect(
            reverse(
                "admin:df_survey_survey_change",
                args=(request.kwargs["survey"].id,),
            )
        )

    def get_export_filename(self, request, queryset, file_format):
        return "%s-%s-%s.%s" % (
            request.kwargs["survey"].title,
            "questions",
            now().strftime("%Y-%m-%d"),
            file_format.get_extension(),
        )


@admin.register(Survey)
class SurveyAdmin(admin.ModelAdmin):
    inlines = [QuestionInline]
    formfield_overrides = {
        JSONField: {"widget": JSONEditor},
    }
    search_fields = ("title",)

    list_display = (
        "id",
        "category",
        "title",
        "description",
        "users_total",
        "users_completed",
    )
    list_filter = ("category__slug",)

    def get_queryset(self, request):
        return super().get_queryset(request).annotate_stats()

    def users_total(self, obj):
        return obj.users_total

    users_total.admin_order_field = "users_total"

    def users_completed(self, obj):
        return obj.users_completed

    users_completed.admin_order_field = "users_completed"

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "<str:survey_id>/export_question_responses/",
                self.admin_site.admin_view(self.export_question_responses_view),
                name="df_survey_survey_export_question_responses",
            ),
            path(
                "<str:survey_id>/import_questions/",
                self.admin_site.admin_view(self.import_questions_view),
                name="df_survey_survey_import_questions",
            ),
            path(
                "<str:survey_id>/process_import_questions/",
                self.admin_site.admin_view(self.process_import_questions_view),
                name="df_survey_survey_process_import_questions",
            ),
            path(
                "<str:survey_id>/export_questions/",
                self.admin_site.admin_view(self.export_questions_view),
                name="df_survey_survey_export_questions",
            ),
            path(
                "<str:survey_id>/assign_users/",
                self.admin_site.admin_view(self.assign_users_view),
                name="df_survey_survey_assign_users",
            ),
        ]
        return custom_urls + urls

    def assign_users_view(self, request, **kwargs):
        self.set_request_kwargs(request, **kwargs)

        if request.method == "POST":
            form = SurveyUsersForm(request.POST)
            if form.is_valid():
                user_surveys = form.save(survey=request.kwargs["survey"])
                if len(user_surveys) > 0:
                    messages.success(
                        request,
                        "Successfully assigned %s users to the survey"
                        % len(user_surveys),
                    )
                else:
                    messages.warning(request, "No users were assigned to the survey")

                return redirect("admin:df_survey_survey_change", kwargs["survey_id"])
        else:
            form = SurveyUsersForm()

        context = {
            "form": form,
            "opts": self.model._meta,
            "survey": Survey.objects.get(id=kwargs["survey_id"]),
        }
        context.update(self.admin_site.each_context(request))
        return render(request, "admin/df_survey/survey/assign_users.html", context)

    def set_request_kwargs(self, request, **kwargs):
        request.kwargs = {"survey": Survey.objects.get(id=kwargs["survey_id"])}

    def import_questions_view(self, request, **kwargs):
        # Redirect to the generic import view with a context tailored for the specific survey
        self.set_request_kwargs(request, **kwargs)
        return QuestionImportExport(Survey, admin.site).import_action(request)

    def process_import_questions_view(self, request, **kwargs):
        # Redirect to the generic import view with a context tailored for the specific survey
        self.set_request_kwargs(request, **kwargs)
        return QuestionImportExport(Survey, admin.site).process_import(request)

    def export_question_responses_view(self, request, **kwargs):
        # Redirect to the generic export view with a context tailored for the specific survey
        self.set_request_kwargs(request, **kwargs)
        return QuestionResponseExport(Survey, admin.site).export_action(request)

    def export_questions_view(self, request, **kwargs):
        # Redirect to the generic export view with a context tailored for the specific survey
        self.set_request_kwargs(request, **kwargs)
        return QuestionImportExport(Survey, admin.site).export_action(request)

    # # TODO: (eugapx) This has to be implemented via importexport
    # def download_results(self, request, survey_template_id):
    #     # Logic to generate and return XLSX file
    #     response = HttpResponse(content_type="application/vnd.ms-excel")
    #     response[
    #         "Content-Disposition"
    #     ] = f'attachment; filename="survey_results_{survey_template_id}.xlsx"'
    #     survey = get_object_or_404(Survey, id=survey_template_id)
    #     wb = openpyxl.Workbook()
    #     ws = wb.active
    #     ws.title = "Survey Results"
    #     for row in survey.get_responses():
    #         ws.append(row)
    #
    #     wb.save(response)
    #     return response

    @admin.action(description="Assign this survey to all users")
    def create_for_all_users(self, request, queryset):
        for survey in queryset:
            UserSurvey.objects.create_for_users(survey=survey, users=None)

    def generate_tasks(self, request, queryset):
        for survey in queryset:
            survey.generate_task()
            survey.save()

    actions = [create_for_all_users, generate_tasks]

    class Media:
        css = {"all": ("df_survey/admin/css/surveys.css",)}


class ResponseInline(ReadOnlyInline):
    model = Response
    extra = 0


class SurveyFilter(AutocompleteFilter):
    title = "Survey"
    field_name = "survey"


@admin.register(UserSurvey)
class UserSurveyAdmin(AdminChangeLinksMixin, admin.ModelAdmin):
    inlines = [ResponseInline]
    formfield_overrides = {
        JSONField: {"widget": JSONEditor},
    }
    autocomplete_fields = ["user", "survey"]
    list_display = (
        "id",
        "survey_link",
        "user",
        "created",
        "modified",
    )
    change_links = ["survey"]
    list_filter = (
        SurveyFilter,
        "survey__category__slug",
        ("result", admin.EmptyFieldListFilter),
    )
    search_fields = ("user__email", "survey__title")

    def parse_survey_response(self, request, queryset):
        for user_survey in queryset:
            user_survey.parse_survey_response()

    def send_notifications(self, request, queryset):
        for user_survey in queryset:
            for notification in UserSurveyNotification.objects.all():
                notification.send(user_survey)

    actions = [parse_survey_response, send_notifications]

    class Media:
        css = {"all": ("df_survey/admin/css/user_surveys.css",)}


@admin.register(Question)
class QuestionAdmin(admin.ModelAdmin):
    list_display = ("question", "type", "format")
    search_fields = ("question",)
    list_filter = ("type",)


@admin.register(Response)
class ResponseAdmin(admin.ModelAdmin):
    list_display = ("usersurvey", "question", "response")
    autocomplete_fields = ("usersurvey", "question")



# 1. There is no point prepending Survey to every model as this is the app name
# 2. Quite a few anti-patterns spotted

from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from df_notifications.decorators import (
    register_reminder_model,
    register_rule_model,
)
from df_notifications.models import (
    NotifiableModelMixin,
    NotificationModelAsyncReminder,
    NotificationModelAsyncRule,
)
from django.contrib.auth import get_user_model
from django.core import exceptions
from django.db import models
from django.db.models import OuterRef, QuerySet, Subquery, Value
from django.db.models.functions import Coalesce
from django.db.models.signals import post_save
from django.dispatch import receiver
from model_utils.models import TimeStampedModel

from df_survey.renderers import SurveyKitRenderer

if TYPE_CHECKING:
    from django.contrib.auth.models import User
else:
    User = get_user_model()


class Category(models.Model):
    slug = models.CharField(max_length=128)

    def __str__(self):
        return self.slug

    class Meta:
        verbose_name_plural = "Survey categories"


def validate_task_json(json):
    # 1. Ensure all target steps are present in rules
    step_identifiers = [step["stepIdentifier"]["id"] for step in json["steps"]]

    errors = []
    if "rules" in json:
        line = 0
        for rule in json["rules"]:
            if rule["type"] == "conditional":
                step_id = rule["triggerStepIdentifier"]["id"]
                if step_id not in step_identifiers:
                    print("VVVs")

                    errors.append(
                        {
                            "task": f"Invalid triggerStepIdentifier '{step_id}' in rule '{line}'"
                        }
                    )
                for k, v in rule["values"].items():
                    if v not in step_identifiers:
                        errors.append(
                            {
                                "task": [
                                    f"Invalid target step '{v}' in value '{k}' of rule '{line}'"
                                ]
                            }
                        )
            line += 1
        if errors:
            raise exceptions.ValidationError(errors)


class SurveyQuerySet(models.QuerySet):
    def annotate_stats(self):
        return self.annotate(
            users_completed=Coalesce(
                Subquery(
                    UserSurvey.objects.filter(
                        survey=OuterRef("pk"), result__isnull=False
                    )
                    .values("survey")
                    .annotate(total=models.Count("pk"))
                    .values("total"),
                    output_field=models.IntegerField(),
                ),
                Value(0),
            ),
            users_total=Coalesce(
                Subquery(
                    UserSurvey.objects.filter(survey=OuterRef("pk"))
                    .values("survey")
                    .annotate(total=models.Count("pk"))
                    .values("total"),
                    output_field=models.IntegerField(),
                ),
                Value(0),
            ),
        )


class Survey(models.Model):
    objects = SurveyQuerySet.as_manager()

    title = models.CharField(max_length=128)
    description = models.TextField(null=True, blank=True)
    sequence = models.PositiveSmallIntegerField(
        help_text="Display sequence, lower means first", default=1000
    )
    category = models.ForeignKey(
        Category, on_delete=models.SET_NULL, null=True, blank=True
    )
    task = models.JSONField(validators=[validate_task_json], null=True, blank=True)

    def generate_task(self):
        self.task = SurveyKitRenderer.generate_task_from_survey(self)

    def get_respondents(self):
        return (
            Response.objects.filter(usersurvey__survey=self)
            .values_list("usersurvey__user__email", flat=True)
            .exclude(usersurvey__user__email="")
            .exclude(usersurvey__user__email__isnull=True)
            .distinct()
        )

    def get_responses_as_csv(self):
        users, responses = self.get_responses_tuple()
        return [["Question", *users], *responses]

    def get_responses(self):
        users = self.get_respondents()
        return self.question_set.all().annotate_responses(users)

    def get_responses_tuple(self):
        users = self.get_respondents()
        qs = self.question_set.all().responses_value_list(users)

        return users, qs

    def __str__(self):
        return f"{self.id}: {self.title}"


class UserSurveyManager(models.Manager):
    def get_queryset(self):
        return super().get_queryset().select_related("survey__category")

    def create_for_users(self, survey=None, users=None):
        if survey:
            users = users or User.objects.filter(is_active=True).exclude(
                id__in=self.filter(survey=survey).values("user_id")
            )
            for user in users:
                self.get_or_create(user=user, survey=survey)


class UserSurvey(TimeStampedModel, NotifiableModelMixin):
    user_attribute = "user"
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    survey = models.ForeignKey(Survey, on_delete=models.CASCADE)
    result = models.JSONField(null=True, blank=True)
    objects = UserSurveyManager()

    # TODO: move to SurveyKit renderer and rename to parse_results
    def pretty_results(self):
        @dataclass
        class ResultEntry:
            step_id: str
            question: str
            answer: str
            answer_full: Any

        results = []
        if self.result:
            questions = {
                q["stepIdentifier"]["id"]: q["title"] for q in self.survey.task["steps"]
            }

            for result in self.result["results"]:
                try:
                    res = result["results"][0]["result"]
                    answer_full = res
                    if isinstance(res, list) and len(res) == 1:
                        res = res[0]

                    if (
                        isinstance(res, dict)
                        and len(res) == 2
                        and "text" in res
                        and "value" in res
                        and res["text"] == res["value"]
                    ):
                        res = res["value"]

                    if isinstance(res, list) and all(
                        isinstance(r, dict) and r.get("text") and r.get("value")
                        for r in res
                    ):
                        res = ", ".join([r["value"] for r in res])

                    step_id = result["id"]["id"]
                    if step_id in questions:
                        results.append(
                            ResultEntry(
                                step_id=step_id,
                                question=questions[step_id],
                                answer=res,
                                answer_full=answer_full,
                            )
                        )
                except LookupError:
                    pass
        return results

    def save(self, *args, **kwargs):
        if self.result is not None:
            self.to_digest = True
        super().save(*args, **kwargs)

    def parse_survey_response(self):
        questions = {q.id: q for q in self.survey.question_set.all()}
        for result in self.pretty_results():
            if result.step_id in questions:
                Response.objects.update_or_create(
                    usersurvey=self,
                    question=questions[result.step_id],
                    defaults={
                        "response": result.answer,
                    },
                )

    def __str__(self):
        return f"{self.user} - {self.survey}"

    class Meta:
        ordering = ["-modified"]
        unique_together = ["user", "survey"]


class QuestionQuerySet(models.QuerySet):
    def annotate_responses(self, users):
        return self.annotate(
            **{
                user: Coalesce(
                    Subquery(
                        queryset=Response.objects.filter(
                            question_id=OuterRef("pk"),
                            usersurvey__user__email=user,
                        ).values_list("response")[:1]
                    ),
                    Value("", output_field=models.TextField()),
                )
                for user in users
            }
        )

    def responses_value_list(self, users):
        return self.annotate_responses(users).values_list("question", *users)


class Question(models.Model):
    objects = QuestionQuerySet.as_manager()

    class Type(models.TextChoices):
        intro = "intro", "Intro"
        completion = "completion", "Completion"
        text = "text", "Text"
        integer = "integer", "Integer"
        date = "date", "Date"
        single = "single", "Single choice"
        multi = "multi", "Multiple choice"

    survey = models.ForeignKey(Survey, on_delete=models.CASCADE)
    sequence = models.PositiveSmallIntegerField(
        help_text="Display sequence, lower means first", default=1000
    )
    question = models.CharField(max_length=255)
    text = models.TextField(default="", blank=True)
    type = models.CharField(choices=Type.choices, max_length=255)
    format = models.TextField(default="", blank=True)

    def __str__(self):
        return self.question

    class Meta:
        ordering = ["sequence"]


class Response(models.Model):
    # TODO csv exporter
    usersurvey = models.ForeignKey(UserSurvey, on_delete=models.CASCADE)
    question = models.ForeignKey(Question, on_delete=models.CASCADE)
    response = models.TextField()

    class Meta:
        unique_together = ["usersurvey", "question"]


@register_rule_model
class UserSurveyNotification(NotificationModelAsyncRule):
    model = UserSurvey
    tracking_fields = []

    def get_users(self, instance: UserSurvey) -> list:
        return [instance.user]


@register_reminder_model
class UserSurveysReminder(NotificationModelAsyncReminder):
    model = UserSurvey

    def get_model_queryset(self) -> QuerySet[UserSurvey]:
        return super().get_model_queryset().filter(result__isnull=True)

    def get_users(self, instance: UserSurvey) -> list:
        return [instance.user]


# TODO: This would make sense if we were doing rewrites
@receiver(post_save, sender=UserSurvey)
def parse_survey_response(sender, instance: UserSurvey, created, **kwargs):
    if instance.result and not instance.response_set.exists():
        instance.parse_survey_response()



from rest_framework import mixins
from rest_framework.viewsets import GenericViewSet

from ..models import UserSurvey
from .serializers import UserSurveyDetailsSerializer, UserSurveySerializer


class UserSurveyViewSet(
    mixins.ListModelMixin,
    mixins.UpdateModelMixin,
    mixins.RetrieveModelMixin,
    GenericViewSet,
):
    serializer_class = UserSurveyDetailsSerializer
    queryset = UserSurvey.objects.all()

    def get_queryset(self):
        return self.queryset.filter(
            user=self.request.user,
            survey__task__isnull=False,
        )

    def get_serializer_class(self):
        if self.action == "list":
            return UserSurveySerializer
        return self.serializer_class



# Generated by Django 4.2.6 on 2023-11-23 08:27

from django.conf import settings
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ("df_survey", "0003_alter_question_options_remove_survey_questions_and_more"),
    ]

    operations = [
        migrations.AlterUniqueTogether(
            name="usersurvey",
            unique_together={("user", "survey")},
        ),
    ]



# Generated by Django 4.2.6 on 2023-11-27 14:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("df_survey", "0005_alter_response_unique_together"),
    ]

    operations = [
        migrations.AddField(
            model_name="question",
            name="text",
            field=models.TextField(blank=True, default=""),
        ),
        migrations.AlterField(
            model_name="question",
            name="type",
            field=models.CharField(
                choices=[
                    ("intro", "Intro"),
                    ("completion", "Completion"),
                    ("text", "Text"),
                    ("integer", "Integer"),
                    ("date", "Date"),
                    ("single", "Single choice"),
                    ("multi", "Multiple choice"),
                ],
                max_length=255,
            ),
        ),
    ]
